from math import sqrt

from fun_dist.distances import point2point
from fun_dist.classes import Point, Circle, LinearFunction


def derivative(f, x):
    dx = 1e-8
    df = f(x + dx) - f(x)
    return df / dx


def func_tangent(f, x_0) -> LinearFunction:
    deriv = derivative(f, x_0)
    k = round(deriv, 5)
    b = round(-deriv * x_0 + f(x_0), 5)
    linear = LinearFunction(k, b)
    return linear


def circle_tangent_len(point: Point, circle: Circle) -> float:
    return sqrt(point2point(point, circle.center) ** 2 - circle.radius ** 2)



from math import sin, cos, sqrt


class Point:
    def __init__(self, x: float, y: float):
        self.x = x
        self.y = y
        self.system_changed = False

    def change_system(self, dx, dy):
        if self.system_changed:
            raise ValueError('System is already changed')
        self.x = self.x - dx
        self.y = self.y - dy
        self.system_changed = True


class Circle:
    def __init__(self, center: Point, radius: float):
        self.center = center
        self.radius = radius


class Ellipse:
    def __init__(self, center: Point, a: float, b: float):
        self.center = center
        self.a = a
        self.b = b

        def ellipse_polar_func(theta):
            return sqrt(1 / ((cos(theta) / a) ** 2 + (sin(theta) / b) ** 2))

        self.polar_func = ellipse_polar_func


class LinearFunction:
    def __init__(self, k, b):
        self.k = k
        self.b = b

    def f(self):
        def func(x):
            return self.k * x + self.b

        return func

    def __str__(self):
        return str(self.k) + 'x + ' + str(self.b)



from fun_dist.distances import *
from fun_dist.classes import *
from fun_dist.tangent import *
from fun_dist.ellipse import point2ellipse

__all__ = ['Point', 'Circle', 'LinearFunction', 'Ellipse',
           'point2point', 'point2func', 'func2func',
           'point2circle', 'point2ellipse', 'func_tangent',
           'circle_tangent_len']



from fun_dist.classes import Point, Circle

from math import sqrt

from scipy.optimize import fmin


def pythagoras(dx, dy):
    """Returns square root (dx^2 + dy^2). Is used for calculating hypotenuse's length using Pythagorean theorem
    :param dx: leg, float
    :param dy: leg, float
    :return: length of the hypotenuse, float
    """
    return sqrt(dx ** 2 + dy ** 2)


def point2point(point_1: Point, point_2: Point):
    """Returns distance between two points in 2D space
    :param point_1: first point coordinates, Point
    :param point_2: second point coordinates, Point
    :return: distance between point_1 and point_2, float
    """
    dx = point_1.x - point_2.x
    dy = point_1.y - point_2.y
    return pythagoras(dx, dy)


class Optimum(object):
    """
    Some function's optimum value and it's coordinates in n-dimensional space
    """

    def __init__(self, distance: float, args: list):
        self.args = args
        self.distance = distance

    def __str__(self):
        return f'distance = {self.distance}\nargs: ({self.args})'


class Distance(object):
    """
    Abstract class with realized find_minimal_dist method
    """

    def dist_func(self, *args):
        # Abstract distance function
        return 0

    def find_minimal_dist(self, *args):
        """
        Finds global minimum of distance function
        :return: Optimum object with minimum value and it's coordinates in n-dimensional space
        """
        init_args, max_iter = args
        min_args = fmin(self.dist_func, init_args, maxiter=max_iter, disp=False)
        return Optimum(self.dist_func(min_args), min_args)


class DistanceBetweenFunctions(Distance):
    def __init__(self, func_a, func_b):
        self.func_a = func_a
        self.func_b = func_b

    def dist_func(self, arguments):
        x_a, x_b = arguments
        point_a = Point(x_a, self.func_a(x_a))
        point_b = Point(x_b, self.func_b(x_b))
        return point2point(point_a, point_b)


def func2func(func_a, func_b, init_args=(0, 0), maxiter=500):
    almost_zero = 1e-4
    dist = DistanceBetweenFunctions(func_a=func_a, func_b=func_b)
    result = dist.find_minimal_dist(init_args, maxiter)
    min_args = result.args
    if abs(func_a(min_args[0]) - func_b(min_args[1])) < almost_zero:
        result.distance = 0
        print('FUNCTIONS INTERSECT')
    return result.distance


class DistanceBetweenPointAndFunction(Distance):
    def __init__(self, point: Point, func):
        self.func = func
        self.point = point

    def dist_func(self, argument):
        x = argument[0]
        point_on_func = Point(x, self.func(x))
        return point2point(self.point, point_on_func)


def point2func(*, point: Point, func, init_args=(0, 0), maxiter=500):
    if func(point.x) == point.y:
        return 0
    dist = DistanceBetweenPointAndFunction(point=point, func=func)
    return dist.find_minimal_dist(init_args, maxiter).distance


def point_in_circle(point: Point, circle: Circle):
    return (point.x - circle.center.x) ** 2 + (point.y - circle.center.x) ** 2 <= circle.radius ** 2


def point2circle(*, point: Point, circle: Circle) -> float:
    if point_in_circle(point, circle):
        print('POINT INSIDE THE CIRCLE')
        return 0
    return point2point(point_1=point, point_2=circle.center) - circle.radius



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="fun_dist",
    version="0.0.3-beta",
    author="Egor Voron",
    author_email="ekvoron01@gmail.com",
    description="A python package, that provides functions for calculating distances and finding tangents of functions",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/EgorVoron/fun_dist",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
    python_requires='>=3.6',
)



from fun_dist.classes import Point, Ellipse

from math import sqrt

from sympy.solvers import solve
from sympy import Symbol


def f(x, a, b):
    return b * sqrt(1 - (x ** 2 / a ** 2))


def D(A: Point, x: float, a: float, b: float):
    return sqrt((x - A.x) ** 2 + (f(x, a, b) - A.y) ** 2)


def get_nearest_point(A: Point, a: float, b: float):
    if A.y < 0:
        A = Point(A.x, -A.y)
    k = 2 - (2 * b ** 2) / (a ** 2)
    p = (2 * b * A.y) / (a ** 2)
    x = Symbol('x')
    answers = solve(-k ** 2 * x ** 4 + 4 * A.x * k * x ** 3 + (
            k ** 2 * a ** 2 - 4 * A.x ** 2 - a ** 2 * p ** 2) * x ** 2 - 4 * A.x * k * x * a ** 2 + 4 * A.x ** 2 * a ** 2,
                    x)
    real_answers = [ans for ans in answers if complex(ans).imag == 0.0]
    positive_answers = [ans for ans in real_answers if ans > 0]
    negative_answers = [ans for ans in real_answers if ans <= 0]
    if A.x > 0:
        ans = min(positive_answers)
    else:
        ans = max(negative_answers)
    return Point(ans, f(ans, a, b))


def point_in_ellipse(point: Point, ellipse: Ellipse):
    return (point.x / ellipse.a) ** 2 + (point.y / ellipse.b) ** 2 <= 1


def point2ellipse(point: Point, ellipse: Ellipse):
    if point2ellipse(point, ellipse):
        print('POINT INSIDE THE ELLIPSE')
        return 0
    point.change_system(dx=ellipse.center.x, dy=ellipse.center.y)
    nearest_point = get_nearest_point(point, ellipse.a, ellipse.b)
    return D(point, nearest_point.x, ellipse.a, ellipse.b)
